"""
verify.py -- verifying solutions for TimPassLib instances

authors: Niels Lindner, Philine Schiewe
website: timpasslib.aalto.fi
"""

from collections import defaultdict
import networkx
import os
import sys

def check_existence(instance_folder, timetable_file = None):
    """check existence of instance files and timetable file"""

    if not os.path.exists(instance_folder):
        sys.exit(f"Error: Directory '{instance_folder}' does not exist.")
    for prefix in ["Activities", "Config", "Events", "OD"]:
        filename = os.path.join(instance_folder, prefix + ".csv")
        if not os.path.exists(filename):
            sys.exit(f"Error: File '{filename}' does not exist.")

    if timetable_file is not None and not os.path.exists(timetable_file):
            sys.exit(f"Error: File '{timetable_file}' does not exist.")

def read_config(instance_folder):
    """read data from Config.csv and write it into a dict"""

    config = {}
    with open(os.path.join(instance_folder, "Config.csv")) as f:
        for line in f:
            if line.startswith("#"):
                continue
            data = [x.strip() for x in line.split(";")]
            config[data[0]] = data[1]
    return config

def create_ean(instance_folder):
    """create event-activity network from Activities.csv and Events.csv file"""

    ean = networkx.DiGraph()
    with open(os.path.join(instance_folder, "Events.csv")) as f:
        for line in f:
            if line.startswith("#"):
                continue
            data = [x.strip() for x in line.split(";")]
            ean.add_node(int(data[0]), type=data[1].replace('"', ''), stop=int(data[2]))
    with open(os.path.join(instance_folder, "Activities.csv")) as f:
        for line in f:
            if line.startswith("#"):
                continue
            data = [x.strip() for x in line.split(";")]
            ean.add_edge(int(data[2]), int(data[3]), index=int(data[0]), type=data[1].replace('"', ''), lower=int(data[4]), upper=int(data[5]), passengers=0)
    return ean

def set_timetable_and_cost(ean, timetable_file, period_length, change_penalty):
    """equip event-activity network with timetable and cost according to change_penalty"""

    with open(timetable_file) as f:
        for line in f:
            if line.startswith("#") or line.strip() == "":
                continue
            data = [x.strip() for x in line.split(";")]
            if len(data) != 2:
                sys.exit(f"Error: Cannot parse line '{line.strip()}'.")
            try:
                ean.nodes[int(data[0])]["timetable"] = int(data[1])
            except KeyError:
                sys.exit(f"Error: There is no event {int(data[0])}.")
            except ValueError:
                sys.exit(f"Error: Cannot parse line '{line.strip()}'.")
    for v in ean.nodes:
        if "timetable" not in ean.nodes[v]:
            sys.exit(f"Error: No time for event {v}.")
    infinity = len(ean.nodes) * max([ean[u][v]["lower"] + period_length + change_penalty for (u, v) in ean.edges])
    for (u, v) in ean.edges:
        tension = ean[u][v]["lower"] + (ean.nodes[v]["timetable"] - ean.nodes[u]["timetable"]- ean[u][v]["lower"]) % period_length
        ean[u][v]["tension"] = tension
        ean[u][v]["cost"] = tension
        if ean[u][v]["type"] == "change":
            ean[u][v]["cost"] += change_penalty
        elif ean[u][v]["type"] in ["sync", "headway"]:
            ean[u][v]["cost"] = infinity
    return ean

def verify_timetable(ean):
    """checks if timetable is feasible on an event-activity network with tensions"""

    feasible = True
    for (u, v) in ean.edges:
        if ean[u][v]["tension"] < ean[u][v]["lower"] or ean[u][v]["tension"] > ean[u][v]["upper"]:
            print(f"The tension {ean[u][v]['tension']} of activity {ean[u][v]['index']} violates the bounds [{ean[u][v]['lower']}, {ean[u][v]['upper']}].")
            feasible = False
    return feasible

def get_od_pairs(instance_folder):
    """read OD matrix from OD.csv file"""

    od = defaultdict(lambda: {})
    with open(os.path.join(instance_folder, "OD.csv")) as f:
        for line in f:
            if line.startswith("#"):
                continue
            data = [x.strip() for x in line.split(";")]
            #od[int(data[0]), int(data[1])] = int(data[2])
            od[int(data[0])][int(data[1])] = int(data[2])
    return od

def count_passengers(od):
    """count the number of passengers"""
    return sum([od[s][t] for s in od for t in od[s]])

def compute_weight(ean, od):
    """compute the total weight on an ean with tensions via shortest path routing of all OD pairs and assign passengers to activities"""

    for v in list(ean.nodes):
        stop = ean.nodes[v]["stop"]
        if ean.nodes[v]["type"] == "departure":
            ean.add_edge(f"super_s_{stop}", v, cost=0)
        elif ean.nodes[v]["type"] == "arrival":
            ean.add_edge(v, f"super_t_{stop}", cost=0)

    i = 0
    total_weight = 0
    for s in od:
        i += 1
        print(f"Routing OD source {i}/{len(od)}...", end='\r', flush=True)
        path = networkx.single_source_dijkstra_path(ean, source=f"super_s_{s}", weight="cost")
        for t in od[s]:
            for (u, v) in zip(path[f"super_t_{t}"][1:-2], path[f"super_t_{t}"][2:-1]):
                ean[u][v]["passengers"] += od[s][t]
    print()
    
    for v in list(ean.nodes):
        if str(v).startswith("super"):
            ean.remove_node(v)
    
    total_weight = 0
    for (u, v) in ean.edges:
        total_weight += ean[u][v]["passengers"] * ean[u][v]["cost"]
    return total_weight

if __name__ == "__main__":
    if len(sys.argv) != 3:
        sys.exit("Usage: python3 verification.py instance_folder timetable_file")

    instance_folder = sys.argv[1]
    timetable_file = sys.argv[2]

    check_existence(instance_folder, timetable_file)
    ean = create_ean(instance_folder)
    od = get_od_pairs(instance_folder)
    config = read_config(instance_folder)
    set_timetable_and_cost(ean, timetable_file, int(config["period_length"]), int(config["ean_change_penalty"]))
    feasible = verify_timetable(ean)
    if feasible:
        print("The timetable is feasible.")
        total_weight = compute_weight(ean, od)
        passengers = count_passengers(od)
        print(f"Objective: {total_weight} ({total_weight/passengers:.2f} per passenger).")
    else:
        print("The timetable is infeasible.")
